local KillingSpree = Rock:NewAddon("KillingSpree", "LibRockDB-1.0", "LibRockComm-1.0", "LibRockEvent-1.0", "LibRockTimer-1.0", "LibRockConfig-1.0", "LibRockConsole-1.0", "LibFeedback-0.1", "LibFuBarPlugin-3.0")
local self = KillingSpree

local _G = _G
local bit = _G.bit
local time = _G.time
local math = _G.math
local string = _G.string
local pairs = _G.pairs
local ipairs = _G.ipairs
local select = _G.select
local tonumber = _G.tonumber
local UIParent = _G.UIParent
local LowHealthFrame = _G.LowHealthFrame
local UIFrameFlash = _G.UIFrameFlash
local GameTooltip = _G.GameTooltip
local PlaySoundFile = _G.PlaySoundFile
local IsAddOnLoaded = _G.IsAddOnLoaded
local IsInInstance = _G.IsInInstance
local UnitName = _G.UnitName
local UnitSex = _G.UnitSex
local UnitFactionGroup = _G.UnitFactionGroup
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetCorpseRecoveryDelay = _G.GetCorpseRecoveryDelay

self:SetCommPrefix("KS2")
self:SetConfigSlashCommand('/killingspree', '/ks')
self:SetDatabase('KillingSpreeDB', 'KillingSpreePCDB')
self:SetFuBarOption('iconPath', "Interface\\AddOns\\KillingSpree\\icon")
self:SetFuBarOption('hasNoColor', true)

function KillingSpree:OnInitialize()
	self.guidregistry = LibStub("LibGUIDRegistry-0.1", true)
	self.sharedmedia = LibStub("LibSharedMedia-3.0", true)
	self.scheme, self.schemes, self.taunts = {}, { "UT2k3", "UT2k4Male", "UT2k4Male2", "UT2k4Female", "UT2k4Female2" }, {
		[1] = {"bam.wav", "BAM!"},
		[2] = {"chickenlol.wav", "LOL /chicken!"},
		[3] = {"crap.wav", "Aww crap, you died!"},
		[4] = {"danceallday.wav", "I can dance all day!"},
		[5] = {"doh.wav", "Doh, you died!"},
		[6] = {"fuckinlag.wav", "Fuckin lag!"},
		[7] = {"haha.wav", "Ha ha, you died!"},
		[8] = {"hahanoobs", "Ha ha, noobs!"},
		[9] = {"headshot.wav", "Boom, headshot!"},
		[10] = {"headshotslow.wav", "Boooom, headshot!"},
		[11] = {"ious.wav", "Its official, you suck!"},
		[12] = {"lawl.wav", "LOL, you died!"},
		[13] = {"noob.wav", "Noob, you died!"},
		[14] = {"nothing.wav", "You own nothing!"},
		[15] = {"pathetic.wav", "That was pathetic!"},
		[16] = {"slap.wav", "/slap, you died!"},
		[17] = {"weak.wav", "You weak pathetic fool!"},
		[18] = {"yeah.wav", "Oh yeah, you died!"},
		[19] = {"youarenothing.wav", "You are nothing!"},
		[20] = {"you likethat.wav", "You like that, dont you!"},
	}
	for scheme = 1, #self.schemes do
		self.scheme[self.schemes[scheme]] = {
			['firstblood'] = { "firstblood.wav", "First Blood!" },
			['headshot'] = { "headshot.wav", "Headshot!" },
			['youhavelost'] = { "youhavelost.wav", "You have lost the match!" },
			['youhavewon'] = { "youhavewon.wav", "You have won the match!" },
			['multi'] = {
				[2] = { "doublekill.wav", "Double Kill!" },
				[3] = { "multikill.wav", "Multi Kill!" },
				[4] = { "megakill.wav", "Mega Kill!" },
				[5] = { "ultrakill.wav", "Ultra Kill!" },
				[6] = { "monsterkill.wav", "Monster Kill!" },
			},
			['spree'] = {
				[10] = { "killingspree.wav", "Killing Spree!" },
				[15] = { "rampage.wav", "Rampage!" },
				[20] = { "dominating.wav", "Dominating!" },
				[25] = { "unstoppable.wav", "Unstoppable!" },
				[30] = { "godlike.wav", "God-like!" },
			},
			['count'] = {
				[1] = { "1.wav", "One!" },
				[2] = { "2.wav", "Two!" },
				[3] = { "3.wav", "Three!" },
				[4] = { "4.wav", "Four!" },
				[5] = { "5.wav", "Five!" },
			},
		}
		if self.schemes[scheme] ~= "UT2k3" then
			self.scheme[self.schemes[scheme]]['humiliation'] = { "humiliation.wav", "Humiliating defeat!" }
			self.scheme[self.schemes[scheme]]['flawless'] = { "flawless.wav", "Flawless victory!" }
			self.scheme[self.schemes[scheme]]['multi'][7] = { "ludicrouskill.wav", "Ludicrous Kill!" }
			self.scheme[self.schemes[scheme]]['multi'][8] = { "holyshit.wav", "Holy Shit!" }
			self.scheme[self.schemes[scheme]]['spree'][35] = { "wickedsick.wav", "Wicked Sick!" }
		end
	end
	if UnitSex('player') == 3 then
		self.defaultscheme = "UT2k4Female"
	else
		self.defaultscheme = "UT2k4Male"
	end
	self:SetDatabaseDefaults('profile', {
		scheme = self.defaultscheme,
		showIcon = true,
		showText = true,
		messages = true,
		suppress = true,
		taunts = true,
		sounds = true,
		visuals = true,
		shared = true,
		slain = true,
--		pvp = true,
		heal = true,
		damage = true,
		s = 300,
		m = 60,
		x = 0,
		y = 150,
		lock = true,
		frame = "feedback",
		fontsize = "20",
		fonttype = "Arial Narrow",
		fontoutline = "OUTLINE",
		fontcolor = {
			r = 1,
			b = 1,
			g = 1,
			a = 1,
		},
	})
	self:SetDatabaseDefaults('char', {
		damage = {},
		heal = {},
		kills = {
			multi = 0,
			spree = 0,
			instance = 0,
		},
	})
	self:SetConfigTable({
		type = 'group',
		name = "Killing Spree",
		desc = "Consecutive kill announcer",
		icon = "Interface\\AddOns\\KillingSpree\\icon",
		args = {
			header0 = {
				order = 0,
				type = 'header',
				name = "commands",
				hidden = function()
					local healrecords, damagerecords = 0, 0
					for record in pairs(self.db.char.heal) do healrecords = healrecords + 1 end
					for record in pairs(self.db.char.damage) do damagerecords = damagerecords + 1 end
					return  healrecords + damagerecords + self.db.char.kills.multi + self.db.char.kills.spree + self.db.char.kills.instance == 0
				end,
			},
			record = {
				order = 1,
				type = 'execute',
				name = "Display records",
				desc = "Displays the damage, heal, multi, spree and instance records",
				func = function()
					if self.db.profile.damage then
						self:Print(string.format("damage records: "))
						if self.db.char.damage then table.sort(self.db.char.damage) end
						for record in pairs(self.db.char.damage) do
							if self.db.char.damage[record].critrate and self.db.char.damage[record].crit.amount and self.db.char.damage[record].hit.amount then
								self:Print(string.format(" %s = %d (%d) [%s\%%]%s", record, self.db.char.damage[record].hit.amount, self.db.char.damage[record].crit.amount, self.db.char.damage[record].critrate, isDoT and " \"DoT\"" or ""))
							elseif self.db.char.damage[record].crit.amount and self.db.char.damage[record].hit.amount then
								self:Print(string.format(" %s = %d (%d)%s", record, self.db.char.damage[record].hit.amount, self.db.char.damage[record].crit.amount, isDoT and " \"DoT\"" or ""))
							elseif self.db.char.damage[record].crit.amount then
								self:Print(string.format(" %s = (%d)%s", record, self.db.char.damage[record].crit.amount, isDoT and " \"DoT\"" or ""))
							elseif self.db.char.damage[record].hit.amount then
								self:Print(string.format(" %s = %d%s", record, self.db.char.damage[record].hit.amount, isDoT and " \"DoT\"" or ""))
							end
						end
					end
					if self.db.profile.heal then
						self:Print(string.format("heal records: "))
						if self.db.char.heal then table.sort(self.db.char.heal) end
						for record in pairs(self.db.char.heal) do
							if self.db.char.heal[record].critrate and self.db.char.heal[record].crit.amount and self.db.char.heal[record].hit.amount then
								self:Print(string.format(" %s = %d (%d) [%s%%]%s", record, self.db.char.heal[record].hit.amount, self.db.char.heal[record].crit.amount, self.db.char.heal[record].critrate, isHoT and " \"HoT\"" or ""))
							elseif self.db.char.heal[record].crit.amount and self.db.char.heal[record].hit.amount then
								self:Print(string.format(" %s = %d (%d)%s", record, self.db.char.heal[record].hit.amount, self.db.char.heal[record].crit.amount, isHoT and " \"HoT\"" or ""))
							elseif self.db.char.heal[record].crit.amount then
								self:Print(string.format(" %s = (%d)%s", record, self.db.char.heal[record].crit.amount, isHoT and " \"HoT\"" or ""))
							elseif self.db.char.heal[record].hit.amount then
								self:Print(string.format(" %s = %d%s", record, self.db.char.heal[record].hit.amount, isHoT and " \"HoT\"" or ""))
							end
						end
					end
					self:Print(string.format("multi record: %s.", self.db.char.kills.multi))
					self:Print(string.format("spree record: %s.", self.db.char.kills.spree))
					self:Print(string.format("instance record: %s.", self.db.char.kills.instance))
				end,
				hidden = function()
					local healrecords, damagerecords = 0, 0
					for record in pairs(self.db.char.heal) do healrecords = healrecords + 1 end
					for record in pairs(self.db.char.damage) do damagerecords = damagerecords + 1 end
					return  healrecords + damagerecords + self.db.char.kills.multi + self.db.char.kills.spree + self.db.char.kills.instance == 0
				end,
			},
			reset = {
				order = 2,
				type = 'execute',
				name = "Reset records",
				desc = "Resets the damage, heal, multi, spree and instance records",
				func = function()
					self.db.char.damage, self.db.char.heal, self.db.char.kills.multi, self.db.char.kills.spree, self.db.char.kills.instance = {}, {}, 0, 0, 0
					self:Print("records reset.")
				end,
				hidden = function()
					local healrecords, damagerecords = 0, 0
					for record in pairs(self.db.char.heal) do healrecords = healrecords + 1 end
					for record in pairs(self.db.char.damage) do damagerecords = damagerecords + 1 end
					return  healrecords + damagerecords + self.db.char.kills.multi + self.db.char.kills.spree + self.db.char.kills.instance == 0
				end,
			},
			header1 = {
				order = 100,
				type = 'header',
				name = "data options",
			},
			scheme = {
				order = 101,
				type = 'choice',
				name = "Sound scheme",
				desc = "Selects the sound scheme",
				get = function()
					return self.db.profile.scheme
				end,
				set = function(scheme)
					self.db.profile.scheme = scheme
					self:Announce("New sound scheme selected", string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['spree'][10][1]))
				end,
				choices = self.schemes,
			},
			s = {
				order = 102,
				type = 'number',
				name = "Spree timer",
				desc = "Sets the spree kill timer",
				get = function()
					return self.db.profile.s
				end,
				set = function(s)
					if s ~= 0 and s < self.db.profile.m then
						self:Print("spree timer must be greater than multi timer or 0")
					else
						self.db.profile.s = s
						self:UpdateVictims()
					end
				end,
				min = 0,
				max = 600,
				step = 10,
			},
			m = {
				order = 103,
				type = 'number',
				name = "Multi timer",
				desc = "Sets the multi kill timer",
				get = function()
					return self.db.profile.m
				end,
				set = function(m)
					if s ~= 0 and m > self.db.profile.s then
						self:Print("multi timer must be less than spree timer if spree timer is not 0")
					else
						self.db.profile.m = m
						self:UpdateVictims()
					end
				end,
				min = 0,
				max = 60,
				step = 1,
			},
			suppress = {
				order = 104,
				type = 'boolean',
				name = "Once only",
				desc = "Toggles repeated announcements",
				get = function()
					return self.db.profile.suppress
				end,
				set = function(suppress)
					self.db.profile.suppress = suppress
				end,
			},
			messages = {
				order = 105,
				type = 'boolean',
				name = "Show messages",
				desc = "Toggles displaying messages",
				get = function()
					return self.db.profile.messages
				end,
				set = function(messages)
					self.db.profile.messages = messages
				end,
			},
			taunts = {
				order = 106,
				type = 'boolean',
				name = "Play taunts",
				desc = "Toggles playing taunts",
				get = function()
					return self.db.profile.taunts
				end,
				set = function(taunts)
					self.db.profile.taunts = taunts
					self:SetListeners()
				end,
			},
			sounds = {
				order = 107,
				type = 'boolean',
				name = "Play sounds",
				desc = "Toggles playing sounds",
				get = function()
					return self.db.profile.sounds
				end,
				set = function(sounds)
					self.db.profile.sounds = sounds
				end,
			},
			visuals = {
				order = 108,
				type = 'boolean',
				name = "Pulse screen",
				desc = "Toggles screen pulsing",
				get = function()
					return self.db.profile.visuals
				end,
				set = function(visuals)
					self.db.profile.visuals = visuals
					if self.db.profile.visuals then
						UIFrameFlash(LowHealthFrame, 0.25, 0.25, 0.5, false, 0, 0)
					end
				end,
			},
			shared = {
				order = 109,
				type = 'boolean',
				name = "Broadcasting",
				desc = "Toggles spree broadcasting",
				get = function()
					return self.db.profile.shared
				end,
				set = function(shared)
					self.db.profile.shared = shared
					self:SetListeners()
				end,
			},
			slain = {
				order = 110,
				type = 'boolean',
				name = "Slain by",
				desc = "Toggles slain by check",
				get = function()
					return self.db.profile.slain
				end,
				set = function(slain)
					self.db.profile.slain = slain
					self:SetListeners()
				end,
			},
--			pvp = {
--				order = 111,
--				type = 'boolean',
--				name = "PvP only",
--				desc = "Toggles PvP check",
--				get = function()
--					return self.db.profile.pvp
--				end,
--				set = function(pvp)
--					self.db.profile.pvp = pvp
--					self:SetListeners()
--				end,
--			},
			heal = {
				order = 112,
				type = 'boolean',
				name = "Record heals",
				desc = "Toggles heal recording",
				get = function()
					return self.db.profile.heal
				end,
				set = function(heal)
					self.db.profile.heal = heal
					self:SetListeners()
				end,
			},
			damage = {
				order = 113,
				type = 'boolean',
				name = "Record damage",
				desc = "Toggles damage recording",
				get = function()
					return self.db.profile.damage
				end,
				set = function(damage)
					self.db.profile.damage = damage
					self:SetListeners()
				end,
			},
			lock = {
				order = 198,
				type = 'boolean',
				name = "Lock frame",
				desc = "Locks the message frame",
				get = function()
					return self.db.profile.lock
				end,
				set = function()
					self.db.profile.lock = not self.db.profile.lock
					if self.db.profile.lock then
						self:LockFeedback("Notification")
					else
						self:UnlockFeedback("Notification")
					end
				end,
				hidden = function()
					return self.db.profile.frame ~= "feedback"
				end,
			},
			frame = {
				order = 199,
				type = 'choice',
				name = "Message frame",
				desc = "Choose the message frame to use",
				get = function()
					return self.db.profile.frame
				end,
				set = function(frame)
					if (frame == "msbt" and IsAddOnLoaded("MikScrollingBattleText")) or (frame == "sct" and IsAddOnLoaded("sct")) or (frame == "parrot" and IsAddOnLoaded("Parrot")) or (frame == "blizzard" and SHOW_COMBAT_TEXT == "1") or frame == "feedback" then
						self.db.profile.frame = frame
						self:Announce("New message frame selected")
					else
						self:Print("frame isnt usable!")
					end
				end,
				choices = {
					["feedback"] = "Feedback",
					["blizzard"] = "Blizzard",
					["parrot"] = "Parrot",
					["msbt"] = "MSBT",
					["sct"] = "SCT",
				},
			},
			header2 = {
				order = 200,
				type = 'header',
				hidden = function()
					return self.db.profile.frame ~= "feedback"
				end,
			},
			fontcolor = {
				order = 201,
				type = 'color',
				name = "Font color",
				desc = "Changes the message font color",
				get = function()
					return self.db.profile.fontcolor.r, self.db.profile.fontcolor.g, self.db.profile.fontcolor.b, self.db.profile.fontcolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.fontcolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
					self:Announce("New font color selected")
				end,
				hasAlpha = function()
					return self.db.profile.frame == "feedback"
				end,
			},
--			fontsize = {
--				order = 202,
--				type = 'choice',
--				name = "Font size",
--				desc = "Changes the message font size",
--				get = function()
--					return self.db.profile.fontsize
--				end,
--				set = function(fontsize)
--					self.db.profile.fontsize = fontsize
--					self:Announce("New font size selected")
--				end,
--				hidden = function()
--					return self.db.profile.frame ~= "feedback"
--				end,
--				choices = {
--					["16"] = "Tiny",
--					["24"] = "Small",
--					["32"] = "Medium",
--					["48"] = "Large",
--				},
--			},
--			fonttype = {
--				order = 203,
--				type = 'choice',
--				name = "Font type",
--				desc = "Changes the message font type",
--				get = function()
--					return self.db.profile.fonttype
--				end,
--				set = function(fonttype)
--					self.db.profile.fonttype = fonttype
--					self:Announce("New font type selected")
--				end,
--				hidden = function()
--					return self.db.profile.frame ~= "feedback"
--				end,
--				choices = self.sharedmedia:List('font'),
--			},
--			fontoutline = {
--				order = 204,
--				type = 'choice',
--				name = "Font outline",
--				desc = "Changes the message font outline",
--				get = function()
--					return self.db.profile.fontoutline
--				end,
--				set = function(fontoutline)
--					self.db.profile.fontoutline = fontoutline
--					self:Announce("New font outline selected")
--				end,
--				hidden = function()
--					return self.db.profile.frame ~= "feedback"
--				end,
--				choices = {
--					[""] = "None",
--					["OUTLINE"] = "Thin",
--					["THICKOUTLINE"] = "Thick",
--				},
--			},
		},
	})
end

function KillingSpree:OnEnable()
	self:CreateFeedback("Notification", { xoffset = self.db.profile.x, yoffset = self.db.profile.y })
	self.factiongroup = UnitFactionGroup('player')
	self.fctdamage = GetCVar("CombatDamage")
	self.fcthealing = GetCVar("CombatHealing")
	self.fctenable = _G.SHOW_COMBAT_TEXT
	self.spawned = time()
	self.users = {}
	self.kills = 0
	self.dead = false
	self:SetVariables()
	self:SetListeners()
end

function KillingSpree:OnDisable()
	SetCVar("CombatDamage", self.fctdamage)
	SetCVar("CombatHealing", self.fcthealing)
	_G.SHOW_COMBAT_TEXT = self.fctenable
	self.db.profile.lock = true
	self.factiongroup = nil
	self.spawned = nil
	self.victims = nil
	self.users = nil
	self.kills = nil
	self.first = nil
	self.multi = nil
	self.spree = nil
	self.instance = nil
	self.lastmulti = nil
	self.lastspree = nil
	self.broadcasted = nil
	self:RemoveAllTimers()
	self:RemoveAllCommListeners()
	self:RemoveAllEventListeners()
end

function KillingSpree:OnUpdateFuBarText()
	local text = self.instance
	if self.db.profile.s > 0 and self.instance ~= self.spree then
		text = text .. ":" .. self.spree
	end
	if self.db.profile.m > 0 and self.spree ~= self.multi then
		text = text .. ":" .. self.multi
	end
	self:SetFuBarText(text)
end

function KillingSpree:OnUpdateFuBarTooltip()
	GameTooltip:AddLine(" Killing Spree ", 1, 1, 0)
	for victim = self.kills, 0, -1 do
		if self.victims[victim] and self.victims[victim].announced then
			if self.db.profile.m > 0 and time() - self.victims[victim].tod < self.db.profile.m then
				if self.db.profile.slain then
					GameTooltip:AddDoubleLine("", "", 0, 0, 0, 0, 0, 0)
					GameTooltip:AddDoubleLine(" " .. self.victims[victim].name .. " ", " " .. self.victims[victim].slayer .. " ", 1, 0, 0, 1, 0, 0);
				else
					GameTooltip:AddLine("", 0, 0, 0)
					GameTooltip:AddLine(" " .. self.victims[victim].name .. " ", 1, 0, 0)
				end
			elseif self.db.profile.s > 0 and time() - self.victims[victim].tod < self.db.profile.s then
				if self.db.profile.slain then
					GameTooltip:AddDoubleLine("", "", 0, 0, 0, 0, 0, 0)
					GameTooltip:AddDoubleLine(" " .. self.victims[victim].name .. " ", " " .. self.victims[victim].slayer .. " ", 1, 1, 0, 1, 1, 0);
				else
					GameTooltip:AddLine("", 0, 0, 0)
					GameTooltip:AddLine(" " .. self.victims[victim].name .. " ", 1, 1, 0)
				end
			else
				if self.db.profile.slain then
					GameTooltip:AddDoubleLine("", "", 0, 0, 0, 0, 0, 0)
					GameTooltip:AddDoubleLine(" " .. self.victims[victim].name .. " ", " " .. self.victims[victim].slayer .. " ", 1, 1, 1, 1, 1, 1);
				else
					GameTooltip:AddLine("", 0, 0, 0)
					GameTooltip:AddLine(" " .. self.victims[victim].name .. " ", 1, 1, 1)
				end
			end
		end
	end
end

function KillingSpree:OnCommReceive(prefix, distribution, sender, message)
	if prefix == "KS1" and sender ~= UnitName('player') then
		if type(message) == 'string' and message == "HELLO" and not self.users[sender] then
			self.users[sender] = true
			self:Print(sender .. " is a Killing Spree user.")
			self:SendCommMessage("GROUP", "HELLO")
		elseif type(message) == 'table' and message.spree then
			self:Announce(string.format("%s is on a killing spree [x%s]", sender, message.spree))
		end
	end
end

function KillingSpree:OnFeedbackFrameMoved(event, frame, xoffset, yoffset)
	self.db.profile.xoffset, self.db.profile.yoffset = xoffset, yoffset
end

function KillingSpree:GetObjectType(objectFlags)
	local COMBATLOG_OBJECT = {
		[1] = { 'guardian', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_TYPE_GUARDIAN) },
		[2] = { 'object', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_TYPE_OBJECT) },
		[3] = { 'pet', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_TYPE_PET) },
		[4] = { 'player', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_TYPE_PLAYER) },
		[5] = { 'target', COMBATLOG_OBJECT_TARGET },
		[6] = { 'focus', COMBATLOG_OBJECT_FOCUS },
		[7] = { 'tank', COMBATLOG_OBJECT_MAINTANK },
		[8] = { 'assist', COMBATLOG_OBJECT_MAINASSIST },
		[9] = { 'star', COMBATLOG_OBJECT_RAIDTARGET1 },
		[10] = { 'circle', COMBATLOG_OBJECT_RAIDTARGET2 },
		[11] = { 'diamond', COMBATLOG_OBJECT_RAIDTARGET3 },
		[12] = { 'triangle', COMBATLOG_OBJECT_RAIDTARGET4 },
		[13] = { 'moon', COMBATLOG_OBJECT_RAIDTARGET5 },
		[14] = { 'square', COMBATLOG_OBJECT_RAIDTARGET6 },
		[15] = { 'cross', COMBATLOG_OBJECT_RAIDTARGET7 },
		[16] = { 'skull', COMBATLOG_OBJECT_RAIDTARGET8 },
		[17] = { 'partypet', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_TYPE_PET) },
		[18] = { 'partyobject', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_TYPE_OBJECT) },
		[19] = { 'partyguardian', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_TYPE_GUARDIAN) },
		[20] = { 'party', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_TYPE_PLAYER) },
		[21] = { 'raidpet', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_RAID, COMBATLOG_OBJECT_TYPE_PET) },
		[22] = { 'raidobject', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_RAID, COMBATLOG_OBJECT_TYPE_OBJECT) },
		[23] = { 'raidguardian', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_RAID, COMBATLOG_OBJECT_TYPE_GUARDIAN) },
		[24] = { 'raid', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_AFFILIATION_RAID, COMBATLOG_OBJECT_TYPE_PLAYER) },
		[25] = { 'friend', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_REACTION_FRIENDLY) },
		[26] = { 'enemy', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_REACTION_HOSTILE) },
		[27] = { 'neutral', bit.bor(COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_REACTION_NEUTRAL) },
		[28] = { 'npc', COMBATLOG_OBJECT_CONTROL_NPC },
	}
	for _, objectInfo in ipairs(COMBATLOG_OBJECT) do
		local objectType, objectTypeFlags = objectInfo[1], objectInfo[2]
		if bit.band(objectFlags, objectTypeFlags) == objectTypeFlags then
			return objectType
		end
	end
	return 'unknown'
end

function KillingSpree:CombatLogEvent(caller, event, timestamp, eventType, sourceGUID, sourceName, sourceFlags, destinationGUID, destinationName, destinationFlags, ...)
	local sourceType, destinationType = self:GetObjectType(sourceFlags), self:GetObjectType(destinationFlags)
	if eventType == "PARTY_KILL" or eventType == "UNIT_DIED" then
		if (self.db.profile.slain and sourceType ~= 'player' and sourceID ~= 'pet' and sourceType ~= 'object' and sourceType ~= 'guardian') then
			return
		else
			self:EnemyDead(eventType, sourceName, sourceType, destinationName, destinationType)
		end
	elseif sourceType == 'player' and eventType == "SWING_DAMAGE" then
	-- amount, school, resisted, blocked, absorbed, critical, glancing, crushing
		local abilityName, amount, isCritical = "Mellee", select(1, ...), select(6, ...)
		self:PlayerDamage(eventType, sourceName, sourceType, destinationName, destinationType, abilityName, amount, isCritical, false)
	elseif sourceType == 'player' and (eventType == "RANGE_DAMAGE" or eventType == "SPELL_DAMAGE" or eventType == "SPELL_PERIODIC_DAMAGE") then
	-- spellId, spellName, spellSchool, amount, school, resisted, blocked, absorbed, critical, glancing, crushing
		local abilityName, amount, isCritical, isDoT = select(2, ...), select(4, ...), select(9, ...), string.find(eventType, "PERIODIC")
		self:PlayerDamage(eventType, sourceName, sourceType, destinationName, destinationType, abilityName, amount, isCritical, isDoT)
	elseif sourceType == 'player' and (eventType == "SPELL_DRAIN" or eventType == "SPELL_LEECH" or eventType == "SPELL_PERIODIC_DRAIN" or eventType == "SPELL_PERIODIC_LEECH") then
	-- spellId, spellName, spellSchool, amount, powerType, extraAmount
		local abilityName, amount, damageType, isDoT = select(2, ...), select(4, ...), select(5, ...), string.find(eventType, "PERIODIC")
		if damageType == -2 then self:PlayerDamage(eventType, sourceName, sourceType, destinationName, destinationType, abilityName, amount, false, isDoT) end
	elseif sourceType == 'player' and (eventType == "SPELL_HEAL" or eventType == "SPELL_PERIODIC_HEAL") then
	-- spellId, spellName, spellSchool, amount, critical
		local abilityName, amount, isCritical, isHoT = select(2, ...), select(4, ...), select(5, ...), string.find(eventType, "PERIODIC")
		self:PlayerHeal(eventType, sourceName, sourceType, destinationName, destinationType, abilityName, amount, isCritical, isHoT)
	end
end

function KillingSpree:BattleGroundEvent(caller, event, message)
	local bgmessage, bgvictory0, bgvictory1, bgstart = string.lower(message), string.lower(VICTORY_TEXT0), string.lower(VICTORY_TEXT1)
	if (self.factiongroup == "Alliance" and string.find(bgmessage, bgvictory1)) or (self.factiongroup == "Horde" and string.find(bgmessage, bgvictory0)) then
		self:Announce(string.format("%s", self.scheme[self.db.profile.scheme]['youhavewon'][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['youhavewon'][1]))
	elseif (self.factiongroup == "Alliance" and string.find(bgmessage, bgvictory0)) or (self.factiongroup == "Horde" and string.find(bgmessage, bgvictory1)) then
		self:Announce(string.format("%s", self.scheme[self.db.profile.scheme]['youhavelost'][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['youhavelost'][1]))
	elseif string.find(bgmessage, "15 seconds") or string.find(bgmessage, "fifteen seconds") then
		self:AddTimer("COUNTDOWN:5", 11, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][5][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][5][1]))
		self:AddTimer("COUNTDOWN:4", 12, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][4][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][4][1]))
		self:AddTimer("COUNTDOWN:3", 13, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][3][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][3][1]))
		self:AddTimer("COUNTDOWN:2", 14, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][2][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][2][1]))
		self:AddTimer("COUNTDOWN:1", 15, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][1][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][1][1]))
	end
end

function KillingSpree:PlayerDead(caller, event)
	if event == "PLAYER_ALIVE" then
		self.dead = false
		self:SetVariables()
	elseif event == "PLAYER_DEAD" and not self.dead then
		self.dead = true
		local taunt = math.random(1, #self.taunts)
		if taunt and self.db.profile.taunts then
			self:Announce(string.format("%s", self.taunts[taunt][2]), string.format("Taunts\\%s", self.taunts[taunt][1]))
		end
	elseif event == "PLAYER_UNGHOST" then
		local resurrection = GetCorpseRecoveryDelay()
		if resurrection and resurrection > 5 then
			self:AddTimer("COUNTDOWN:5", resurrection - 5, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][5][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][5][1]))
			self:AddTimer("COUNTDOWN:4", resurrection - 4, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][4][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][4][1]))
			self:AddTimer("COUNTDOWN:3", resurrection - 3, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][3][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][3][1]))
			self:AddTimer("COUNTDOWN:2", resurrection - 2, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][2][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][2][1]))
			self:AddTimer("COUNTDOWN:1", resurrection - 1, "Announce", string.format("%s", self.scheme[self.db.profile.scheme]['count'][1][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['count'][1][1]))
		end	
	end
end

function KillingSpree:EnemyDead(eventType, sourceName, sourceType, destinationName, destinationType)
	if sourceName and destinationName then
		if not self.victims[self.kills] then
			self.victims[self.kills] = {}
		end
		self.victims[self.kills].tod = time()
		self.victims[self.kills].name = destinationName
		if sourceType == 'player' or sourceType == 'pet' or sourceType == 'object' or sourceType == 'guardian' then
			self.victims[self.kills].slayer = "You"
		else
			self.victims[self.kills].slayer = sourceName
		end
		if not self.victims[self.kills].announced then
			self.victims[self.kills].announced = true
			self:UpdateVictims()
			self.kills = self.kills + 1
			if self.db.profile.s == 0 then
				self.spree = self.instance
			end
			if self.db.profile.m == 0 then
				self.multi = self.spree
			end
			if self.multi > self.db.char.kills.multi then
				self.db.char.kills.multi = self.multi
			end
			if self.spree > self.db.char.kills.spree then
				self.db.char.kills.spree = self.spree
			end
			if self.instance > self.db.char.kills.instance then
				self.db.char.kills.instance = self.instance
			end
			if self.instance > 1 and self.scheme[self.db.profile.scheme]['multi'][self.multi] and (not self.db.profile.suppress or self.lastmulti < self.multi) then
				self.lastmulti = self.multi
				self:Announce(string.format("%s [x%d]", self.scheme[self.db.profile.scheme]['multi'][self.multi][2], self.multi), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['multi'][self.multi][1]), true)
			elseif self.instance > 1 and self.scheme[self.db.profile.scheme]['spree'][self.spree] and (not self.db.profile.suppress or self.lastspree < self.spree) then
				self.lastspree = self.spree
				if self.db.profile.shared and (not self.db.profile.suppress or self.broadcasted < self.spree) then
					self.broadcasted = self.spree
					self:SendCommMessage("GROUP", { multi = self.multi, spree = self.spree, instance = self.instance, m = self.db.profile.m, s = self.db.profile.s, t = time() - self.spawned, slain = self.db.profile.slain })
				end
				self:Announce(string.format("%s [x%d]", self.scheme[self.db.profile.scheme]['spree'][self.spree][2], self.spree), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['spree'][self.spree][1]), true)
			elseif self.instance == 1 and self.scheme[self.db.profile.scheme]['firstblood'] then
				self:Announce(string.format("%s", self.scheme[self.db.profile.scheme]['firstblood'][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['firstblood'][1]), true)
			elseif self.scheme[self.db.profile.scheme]['headshot'] then
				self:Announce(string.format("%s", self.scheme[self.db.profile.scheme]['headshot'][2]), string.format("Interface\\AddOns\\KillingSpree\\snds\\%s\\%s", self.db.profile.scheme, self.scheme[self.db.profile.scheme]['headshot'][1]), true)
			end
		end
	end
end

function KillingSpree:PlayerHeal(eventType, sourceName, sourceType, destinationName, destinationType, abilityName, amount, isCritical, isHoT)
	if not self.db.char.heal then self.db.char.heal = {} end
	if not self.db.char.heal[abilityName] then
		self.db.char.heal[abilityName] = {}
		self.db.char.heal[abilityName].hit = {}
		self.db.char.heal[abilityName].crit = {}
		self.db.char.heal[abilityName].hits = 0
		self.db.char.heal[abilityName].crits = 0
		self.db.char.heal[abilityName].isHoT = isHoT
	end
	if isCritical then
		self.db.char.heal[abilityName].crits = self.db.char.heal[abilityName].crits + 1
		if not self.db.char.heal[abilityName].crit.amount or self.db.char.heal[abilityName].crit.amount < amount then
			self.db.char.heal[abilityName].crit.ltime = time()
			self.db.char.heal[abilityName].crit.location = GetRealZoneText()
			self.db.char.heal[abilityName].crit.destination = destinationName
			self.db.char.heal[abilityName].crit.amount = amount
			self:Announce(string.format("New %s record crit : %d%s", abilityName, amount, isHoT and " (HoT)" or ""), "Sound\\Spells\\LevelUp.wav", true)
		end
	else
		self.db.char.heal[abilityName].hits = self.db.char.heal[abilityName].hits + 1
		if not self.db.char.heal[abilityName].hit.amount or self.db.char.heal[abilityName].hit.amount < amount then
			self.db.char.heal[abilityName].hit.ltime = time()
			self.db.char.heal[abilityName].hit.location = GetRealZoneText()
			self.db.char.heal[abilityName].hit.destination = destinationName
			self.db.char.heal[abilityName].hit.amount = amount
			self:Announce(string.format("New %s record hit : %d%s", abilityName, amount, isHoT and " (HoT)" or ""), "Sound\\Spells\\LevelUp.wav", true)
		end
	end
	if self.db.char.heal[abilityName].crits > 0 and self.db.char.heal[abilityName].hits > 0 then
		self.db.char.heal[abilityName].critrate = floor((self.db.char.heal[abilityName].crits / (self.db.char.heal[abilityName].crits + self.db.char.heal[abilityName].hits)) * 100)
	end
end

function KillingSpree:PlayerDamage(eventType, sourceName, sourceType, destinationName, destinationType, abilityName, amount, isCritical, isDoT)
	if not self.db.char.damage then self.db.char.damage = {} end
	if not self.db.char.damage[abilityName] then
		self.db.char.damage[abilityName] = {}
		self.db.char.damage[abilityName].hit = {}
		self.db.char.damage[abilityName].crit = {}
		self.db.char.damage[abilityName].hits = 0
		self.db.char.damage[abilityName].crits = 0
		self.db.char.damage[abilityName].isDoT = isDoT
	end
	if isCritical then
		self.db.char.damage[abilityName].crits = self.db.char.damage[abilityName].crits + 1
		if not self.db.char.damage[abilityName].crit.amount or self.db.char.damage[abilityName].crit.amount < amount then
			self.db.char.damage[abilityName].crit.ltime = time()
			self.db.char.damage[abilityName].crit.location = GetRealZoneText()
			self.db.char.damage[abilityName].crit.destination = destinationName
			self.db.char.damage[abilityName].crit.amount = amount
			self:Announce(string.format("New %s record crit : %d%s", abilityName, amount, isDoT and " (DoT)" or ""), "Sound\\Spells\\LevelUp.wav", true)
		end
	else
		self.db.char.damage[abilityName].hits = self.db.char.damage[abilityName].hits + 1
		if not self.db.char.damage[abilityName].hit.amount or self.db.char.damage[abilityName].hit.amount < amount then
			self.db.char.damage[abilityName].hit.ltime = time()
			self.db.char.damage[abilityName].hit.location = GetRealZoneText()
			self.db.char.damage[abilityName].hit.destination = destinationName
			self.db.char.damage[abilityName].hit.amount = amount
			self:Announce(string.format("New %s record hit : %d%s", abilityName, amount, isDoT and " (DoT)" or ""), "Sound\\Spells\\LevelUp.wav", true)
		end
	end
	if self.db.char.damage[abilityName].crits > 0 and self.db.char.damage[abilityName].hits > 0 then
		self.db.char.damage[abilityName].critrate = floor((self.db.char.damage[abilityName].crits / (self.db.char.damage[abilityName].crits + self.db.char.damage[abilityName].hits)) * 100)
	end
end

function KillingSpree:Announce(message, sound, visual)
	if message and self.db.profile.messages then
		if self.db.profile.frame == "sct" and IsAddOnLoaded("sct") then
			SCT:DisplayMessage(message, self.db.profile.fontcolor)
		elseif self.db.profile.frame == "parrot" and IsAddOnLoaded("Parrot") then
			Parrot:ShowMessage(message, "Notification", false, self.db.profile.fontcolor.r, self.db.profile.fontcolor.g, self.db.profile.fontcolor.b)
		elseif self.db.profile.frame == "msbt" and IsAddOnLoaded("MikScrollingBattleText") then
			MikSBT.DisplayMessage(message, MikSBT.DISPLAYTYPE_NOTIFICATION, false, self.db.profile.fontcolor.r, self.db.profile.fontcolor.g, self.db.profile.fontcolor.b)
		elseif self.db.profile.frame == "blizzard" and SHOW_COMBAT_TEXT == "1" then
			CombatText_AddMessage(message, COMBAT_TEXT_SCROLL_FUNCTION, self.db.profile.fontcolor.r, self.db.profile.fontcolor.g, self.db.profile.fontcolor.b)
		else
			local data = {
				text = message,
				fontcolor = self.db.profile.fontcolor,
			}
			if sound and self.db.profile.sounds then
				data.sound = sound
			end
			if visual and self.db.profile.visuals then
				data.visual = function() UIFrameFlash(LowHealthFrame, 0.25, 0.25, 0.5, false, 0, 0) end
			end
			self:AddFeedbackData("Notification", data)
			return
		end
	end
	if sound and self.db.profile.sounds then
		PlaySoundFile(sound)
	end
	if visual and self.db.profile.visuals then
		UIFrameFlash(LowHealthFrame, 0.25, 0.25, 0.5, false, 0, 0)
	end
end

function KillingSpree:UpdateUsers()
	if GetNumRaidMembers() > 1 or GetNumRaidMembers() > 1 then
		self:SendCommMessage("GROUP", "HELLO")
	else
		self.users = {}
	end
end

function KillingSpree:UpdateVictims()
	local multi, spree, instance, refresh = 0, 0, 0
	for victim = self.kills, 0, -1 do
		if self.victims[victim] and self.victims[victim].announced then
			local difference = time() - self.victims[victim].tod
			if self.db.profile.m > 0 and difference < self.db.profile.m then
				if not refresh or refresh > self.db.profile.m - difference then
					refresh = self.db.profile.m - difference
				end
				multi = multi + 1
			elseif self.db.profile.s > 0 and difference < self.db.profile.s then
				if not refresh or refresh >  self.db.profile.s - difference then
					refresh = self.db.profile.s - difference
				end
				spree = spree + 1
			else
				instance = instance + 1
			end
		end
	end
	if self.multi ~= multi or self.spree ~= spree + multi or self.instance ~= instance then
		self.instance = instance + spree + multi
		self.spree = spree + multi
		self.multi = multi
		self:UpdateFuBarPlugin()
	end
	if refresh then
		self:RemoveTimer("REFRESH")
		self:AddTimer("REFRESH", refresh, "UpdateVictims")
	end
end

function KillingSpree:SetVariables()
	self.instanceType = select(2, IsInInstance())
	self.victims = {}
	self.multi = 0
	self.spree = 0
	self.instance = 0
	self.lastmulti = 0
	self.lastspree = 0
	self.broadcasted = 0
	self:RemoveAllTimers()
	self:UpdateFuBarPlugin()
end

function KillingSpree:SetListeners()
	self:RemoveAllCommListeners()
	self:RemoveAllEventListeners()
	if self.db.profile.shared then
		self:AddCommListener("KS1", "GROUP", "OnCommReceive")
		self:AddEventListener("RAID_ROSTER_UPDATE", "UpdateUsers")
		self:AddEventListener("PARTY_MEMBERS_CHANGED", "UpdateUsers")
	end
	self:AddCallback("OnFeedbackFrameMoved")
	self:AddEventListener("PLAYER_DEAD", "PlayerDead")
	self:AddEventListener("PLAYER_ALIVE", "PlayerDead")
	self:AddEventListener("PLAYER_UNGHOST", "PlayerDead")
	self:AddEventListener("ZONE_CHANGED_NEW_AREA", "SetListeners")
	self:AddEventListener("CHAT_MSG_BG_SYSTEM_NEUTRAL", "BattleGroundEvent")
	self:AddEventListener("COMBAT_LOG_EVENT_UNFILTERED", "CombatLogEvent")
end
